import uuid

import barker.schemas.printer as schemas

from fastapi import APIRouter, Depends, HTTPException, Security, status
from sqlalchemy import select
from sqlalchemy.exc import SQLAlchemyError

from ..core.security import get_current_active_user as get_user
from ..db.session import SessionFuture
from ..models.printer import Printer
from ..schemas.user_token import UserToken


router = APIRouter()


@router.post("", response_model=schemas.Printer)
def save(
    data: schemas.PrinterIn,
    user: UserToken = Security(get_user, scopes=["printers"]),
) -> schemas.Printer:
    try:
        with SessionFuture() as db:
            item = Printer(
                name=data.name,
                address=data.address,
                cut_code=bytes.fromhex(data.cut_code).decode(),
            )
            db.add(item)
            db.commit()
            return printer_info(item)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.put("/{id_}", response_model=schemas.Printer)
def update_route(
    id_: uuid.UUID,
    data: schemas.PrinterIn,
    user: UserToken = Security(get_user, scopes=["printers"]),
) -> schemas.Printer:
    try:
        with SessionFuture() as db:
            item: Printer = db.execute(select(Printer).where(Printer.id == id_)).scalar_one()
            item.name = data.name
            item.address = data.address
            item.cut_code = bytearray.fromhex(data.cut_code).decode()
            db.commit()
            return printer_info(item)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.delete("/{id_}", response_model=schemas.PrinterBlank)
def delete_route(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["printers"]),
) -> schemas.PrinterBlank:
    try:
        with SessionFuture() as db:
            item: Printer = db.execute(select(Printer).where(Printer.id == id_)).scalar_one()
            db.delete(item)
            db.commit()
            return printer_blank()
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.get("", response_model=schemas.PrinterBlank)
def show_blank(
    user: UserToken = Security(get_user, scopes=["printers"]),
) -> schemas.PrinterBlank:
    return printer_blank()


@router.get("/list", response_model=list[schemas.Printer])
def show_list(user: UserToken = Depends(get_user)):
    with SessionFuture() as db:
        return [printer_info(item) for item in db.execute(select(Printer).order_by(Printer.name)).scalars().all()]


@router.get("/{id_}", response_model=schemas.Printer)
def show_id(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["printers"]),
) -> schemas.Printer:
    with SessionFuture() as db:
        item: Printer = db.execute(select(Printer).where(Printer.id == id_)).scalar_one()
        return printer_info(item)


def printer_info(item: Printer) -> schemas.Printer:
    return schemas.Printer(
        id_=item.id,
        name=item.name,
        address=item.address,
        cut_code=bytes(item.cut_code, "ascii").hex(),
    )


def printer_blank() -> schemas.PrinterBlank:
    return schemas.PrinterBlank(name="", address="", cut_code="1d564103")
